////////////////////////////////////////////////////////////////////////
// $Id: CalWireArgoneut.cxx,v 1.36 2010/09/15  bpage Exp $
//
// CalWireArgoneut class
//
// brebel@fnal.gov, greenlee@fnal.gov, tjyang@fnal.gov, baller@fnal.gov
//
////////////////////////////////////////////////////////////////////////
#ifndef CALWIRET962_H
#define CALWIRET962_H

#include <stdint.h>

// Framework includes
#include "messagefacility/MessageLogger/MessageLogger.h" 
#include "fhiclcpp/ParameterSet.h" 
#include "cetlib_except/exception.h"
#include "cetlib/search_path.h"
#include "art/Framework/Core/ModuleMacros.h" 
#include "art/Framework/Principal/Event.h" 
#include "art/Framework/Core/EDProducer.h"
#include "art/Framework/Principal/Handle.h" 
#include "canvas/Persistency/Common/Ptr.h" 
#include "canvas/Persistency/Common/Assns.h" 
#include "art/Framework/Services/Registry/ServiceHandle.h" 
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Services/Optional/TFileDirectory.h"

// LArSoft includes
#include "larcoreobj/SimpleTypesAndConstants/RawTypes.h" // raw::ChannelID_t
#include "larcore/Geometry/Geometry.h"
#include "larevt/CalibrationDBI/Interface/ChannelStatusService.h"
#include "larevt/CalibrationDBI/Interface/ChannelStatusProvider.h"
#include "lardataobj/RawData/RawDigit.h"
#include "lardataobj/RawData/raw.h" // raw::Uncompress()
#include "lardataobj/RecoBase/Wire.h"
#include "lardata/ArtDataHelper/WireCreator.h"
#include "lardata/Utilities/SignalShaping.h"
#include "lardata/Utilities/AssociationUtil.h"
// ArgoNeuT includes
#include "T962Utilities/SignalShapingServiceT962.h"

// ROOT includes
#include "TFile.h"
#include "TH1D.h"

///creation of calibrated signals on wires
namespace caldata {

  class CalWireArgoneut : public art::EDProducer {

  public:
    
    // create calibrated signals on wires. this class runs 
    // an fft to remove the electronics shaping.     
    explicit CalWireArgoneut(fhicl::ParameterSet const& pset); 
    virtual ~CalWireArgoneut();
    
    void produce(art::Event& evt); 
    void beginJob(); 
    void endJob();                 
    void reconfigure(fhicl::ParameterSet const& p);
 
  private:
    
    std::vector<int> fBaseSampleBins;        ///< number of postsample bins
    std::string  fDigitModuleLabel;  ///< module that made digits
    bool  fDodQdxCalib;          ///< Do we apply wire-by-wire calibration?
    std::map<unsigned int, float> fdQdxCalib; ///< Map to do wire-by-wire calibration, key is channel number, content is correction factor
    
    void SubtractBaseline(std::vector<float>& holder, int baseSampleBins);

    std::vector<int> fHistoChannels;
    TH1D* fRawNoise;        ///< distribution of noise counts for a selected channel in one event
    TH1D* fDeconNoise;      
    
  }; // class CalWireArgoneut
}

namespace caldata{

  //-------------------------------------------------
  CalWireArgoneut::CalWireArgoneut(fhicl::ParameterSet const& pset)
  {
    this->reconfigure(pset);

    produces< std::vector<recob::Wire> >();
    produces<art::Assns<raw::RawDigit, recob::Wire>>();
    
  }
  
  //-------------------------------------------------
  CalWireArgoneut::~CalWireArgoneut()
  {
  }

  //////////////////////////////////////////////////////
  void CalWireArgoneut::reconfigure(fhicl::ParameterSet const& p)
  {
    fDigitModuleLabel = p.get<std::string>("DigitModuleLabel", "daq");
    fBaseSampleBins   = p.get<std::vector<int>>("BaseSampleBins");
    if(fBaseSampleBins.size() != 2) {
      std::cout<<"BaseSample bins must be size 2\n";
      exit(1);
    }
    fDodQdxCalib      = p.get<bool>("DodQdxCalib", false);
    if (fDodQdxCalib){
      std::string fileName = p.get< std::string >("dQdxCalibFileName");
      std::string fullname;
      cet::search_path sp("FW_SEARCH_PATH");
      sp.find_file(fileName, fullname);
      
      if (fullname.empty()) {
        std::cout << "Input file " << fileName << " not found" << std::endl;
        throw cet::exception("File not found");
      }
      else
        std::cout << "Applying wire-by-wire calibration using file " << fileName << std::endl;
      
      std::ifstream inFile(fullname, std::ios::in);
      std::string line;
      
      while (std::getline(inFile,line)) {
        unsigned int channel;
        float        constant;
        std::stringstream linestream(line);
        linestream >> channel >> constant;
        fdQdxCalib[channel] = constant;
        if (channel%100==0) std::cout<<"Channel "<<channel<<" correction factor "<<fdQdxCalib[channel]<<std::endl;
      }
    }
    fHistoChannels.resize(0);
    if(p.has_key("HistoChannels")) fHistoChannels = p.get<std::vector<int>>("HistoChannels"); // histogram wire signal on one channel
  } // reconfigure

  //-------------------------------------------------
  void CalWireArgoneut::beginJob()
  {  
    if(!fHistoChannels.empty()) {
      art::ServiceHandle<art::TFileService> tfs;
      fRawNoise = tfs->make<TH1D>("RawNoise", "Raw Noise (ADC)", 1000, -10., 10.);
      fDeconNoise = tfs->make<TH1D>("DeconNoise", "Decon Noise (ADC)", 1000, -10., 10.);    }
  } // beginJob

  //////////////////////////////////////////////////////
  void CalWireArgoneut::endJob()
  {  
  }
  
  //////////////////////////////////////////////////////
  void CalWireArgoneut::produce(art::Event& evt)
  {
    
      
    art::ServiceHandle<geo::Geometry> geom;
    art::ServiceHandle<util::LArFFT> fFFT;
    art::ServiceHandle<util::SignalShapingServiceT962> sss;
    int transformSize = fFFT->FFTSize();
    lariov::ChannelStatusProvider const& channelStatus = art::ServiceHandle<lariov::ChannelStatusService>()->GetProvider();
    // make a collection of Wires
    std::unique_ptr<std::vector<recob::Wire> > wirecol(new std::vector<recob::Wire>);
    // ... and an association set
    std::unique_ptr<art::Assns<raw::RawDigit,recob::Wire> > WireDigitAssn(new art::Assns<raw::RawDigit,recob::Wire>);
    
    // Read in the digit List object(s). 
    art::Handle< std::vector<raw::RawDigit> > digitVecHandle;
    evt.getByLabel(fDigitModuleLabel, digitVecHandle);
    if (!digitVecHandle->size()) return;

    // Use the handle to get a particular (0th) element of collection to find the data size
    art::Ptr<raw::RawDigit> digitVec0(digitVecHandle, 0);
    unsigned int dataSize = digitVec0->Samples(); //size of raw data vectors
    for(unsigned short plane = 0; plane < 2; ++plane) {
      if(fBaseSampleBins[plane] > 0 && dataSize % fBaseSampleBins[plane] != 0) {
        std::cout<<"Set BaseSampleBins modulo dataSize= "<<dataSize<<"\n";
        exit(1);
      }
    } // plane
    
    // loop over all wires
    for(unsigned int rdIter = 0; rdIter < digitVecHandle->size(); ++rdIter) {
      art::Ptr<raw::RawDigit> digitVec(digitVecHandle, rdIter);
      raw::RawDigit::ADCvector_t rawadc(digitVec->Samples());  // vector holding uncompressed adc values
      raw::Uncompress(digitVec->ADCs(), rawadc, digitVec->Compression());
      uint32_t channel = digitVec->Channel();
      if(channelStatus.IsBad(channel)) continue;
      float ped = digitVec->GetPedestal();
      std::vector<float> holder(transformSize, 0.);
      for(unsigned short bin = 0; bin < dataSize; ++bin) holder[bin] = rawadc[bin] - ped;
      if(!fHistoChannels.empty() && (int)channel >= fHistoChannels[0] && (int)channel <= fHistoChannels[1]) {
        // fill the noise histogram on the first occurrence of the selected channel
        for(unsigned short bin = 0; bin < dataSize; ++bin) fRawNoise->Fill(holder[bin]);
      } // Histo Channels
      sss->Deconvolute(channel, holder);
      // throw away the non-physical section
      holder.resize(dataSize);
      // adaptive baseline subtraction in the collection plane
      unsigned short plane = 0;
      if(channel >= 240) plane = 1;
      SubtractBaseline(holder, fBaseSampleBins[plane]);
      if(!fHistoChannels.empty() && (int)channel >= fHistoChannels[0] && (int)channel <= fHistoChannels[1]) {
        // fill the noise histogram on the first occurrence of the selected channel
        for(unsigned short bin = 0; bin < dataSize; ++bin) fDeconNoise->Fill(holder[bin]);
      } // Histo Channels

      // apply wire-by-wire calibration
      if (fDodQdxCalib){
        if(fdQdxCalib.find(channel) != fdQdxCalib.end()){
          float constant = fdQdxCalib[channel];
          //std::cout<<channel<<" "<<constant<<std::endl;
          for (unsigned short bin = 0; bin < holder.size(); ++bin) holder[bin] *= constant;
        } // fdQdxCalib.find(channel) != fdQdxCalib.end()
      } // fDodQdxCalib
      // Make a single ROI that spans the entire data size
      wirecol->push_back(recob::WireCreator(holder,*digitVec).move());
      if (!util::CreateAssn(*this, evt, *wirecol, digitVec, *WireDigitAssn)) {
        throw art::Exception(art::errors::ProductRegistrationFailure)
        << "Can't associate wire #" << (wirecol->size() - 1)
        << " with raw digit #" << digitVec.key();
      } // if failed to add association
    } // rdIter

    if(wirecol->size() == 0)
      mf::LogWarning("CalWireArgoneut") << "No wires made for this event.";
    
    evt.put(std::move(wirecol));
    evt.put(std::move(WireDigitAssn));
    
    // turn off histograms
    if(!fHistoChannels.empty()) fHistoChannels.resize(0);
    
    return;
  } // produce
  
  //////////////////////////////////////////////////////
  void CalWireArgoneut::SubtractBaseline(std::vector<float>& holder, int baseSampleBins)
  {
    // subtract baseline using linear interpolation between regions defined
    // by the datasize and baseSampleBins
    if(baseSampleBins <= 0) return;
    constexpr double fBaseVarCut = 20;
    // number of points to characterize the baseline
    unsigned short nBasePts = holder.size() / baseSampleBins;
    // the baseline offset vector
    std::vector<double> base(nBasePts, 0.);
    // find the average value in each region, using values that are
    // similar
    double fbins = baseSampleBins;
    unsigned short nfilld = 0;
    for(unsigned short ii = 0; ii < nBasePts; ++ii) {
      unsigned short loBin = ii * baseSampleBins;
      unsigned short hiBin = loBin + baseSampleBins;
      double ave = 0.;
      double sum = 0.;
      for(unsigned short bin = loBin; bin < hiBin; ++bin) {
        ave += holder[bin];
        sum += holder[bin] * holder[bin];
      } // bin
      ave = ave / fbins;
      double var = (sum - fbins * ave * ave) / (fbins - 1.);
      // Set the baseline for this region if the variance is small
      if(var < fBaseVarCut) {
        base[ii] = ave;
        ++nfilld;
      }
    } // ii

    // interpolate and subtract
    // find the first non-zero sample
    unsigned short thisRegion = 0;
    unsigned short nextRegion = 1;
    for(nextRegion = 1; nextRegion < base.size(); ++nextRegion) if(base[nextRegion] != 0) break;
    if(nextRegion == base.size()) return;
    double nBins = (nextRegion - thisRegion) * baseSampleBins;
    double slp = (base[nextRegion] - base[thisRegion]) / nBins;
    for(unsigned short bin = 0; bin < holder.size(); ++bin) {
      // determine which region we are in
      short region = bin / baseSampleBins;
      if(region == nextRegion) {
        // update the slope for the next region
        unsigned short nnr = 0;
        for(nnr = nextRegion + 1; nnr < base.size(); ++nnr) if(base[nnr] != 0) break;
        if(nnr == base.size()) nnr = base.size() - 1;
        thisRegion = nextRegion;
        nextRegion = nnr;
        if(thisRegion == nextRegion) return;
        nBins = (nextRegion - thisRegion) * baseSampleBins;
        slp = (base[nextRegion] - base[thisRegion]) / nBins;
      }
      double bs = base[thisRegion] + (double)(bin - thisRegion * baseSampleBins) * slp;
      holder[bin] -= bs;
    } // bin
  } // SubtractBaseline
  
} // end namespace caldata

namespace caldata{

  DEFINE_ART_MODULE(CalWireArgoneut)
  
} // end namespace caldata

#endif // CALWIRET962_H


